import typing

class VLPError(Exception): ...

class VLPTypeError(VLPError):
    def __init__(self, value, *typeargv: type) -> None: ...

class VLPSequenceLenError(VLPError):
    def __init__(self, value: typing.Sequence, length: int) -> None: ...

class VLPValueError(VLPError):
    def __init__(self, value, *argv) -> None: ...

class VLPValueRangeError(VLPError):
    def __init__(self, value, low, hignt) -> None: ...



def draw_position_contrast(pred, true, title: str = ..., xlabel: str = ..., ylabel: str = ..., zlabel: str = ..., savepath=..., showfig: bool = ...) -> None: ...
def draw_history(history, title: str = ..., savepath=..., showfig: bool = ...) -> None: ...
def draw_error_hist(error, bins: int = ..., rwidth: float = ..., title: str = ..., savepath=..., showfig: bool = ...): ...
def cdfplot(error, title: str = ..., savepath=..., showfig: bool = ...): ...



import typing
from ..error import VLPValueError as VLPValueError
from _typeshed import Incomplete

class BaseFilter:
    Fc: Incomplete
    Fs: Incomplete
    Wp: Incomplete
    Ws: Incomplete
    btype: Incomplete
    analog: Incomplete
    def __init__(self, fc, fs, D: list = ..., G: list = ..., btype: str = ..., analog: bool = ..., *args, **kwargs) -> None: ...
    def show_wave(self, N, whole: bool = ..., savepath=..., showfig: bool = ...) -> None: ...
    def filter(self, x, lfilter: bool = ...): ...

class ButterFilter(BaseFilter):
    def __init__(self, fc, fs, D: list = ..., G: list = ..., btype: str = ..., analog: bool = ..., *args, **kwargs) -> None: ...

class ChebyFilter(BaseFilter):
    def __init__(self, fc, fs, D: list = ..., G: list = ..., btype: str = ..., analog: bool = ..., cheby: int = ..., *args, **kwargs) -> None: ...

class EllipFilter(BaseFilter):
    def __init__(self, fc, fs, D: list = ..., G: list = ..., btype: str = ..., analog: bool = ..., *args, **kwargs) -> None: ...

class BesselFilter(BaseFilter):
    def __init__(self, fc, fs, n, Wn: typing.Union[int, list], btype: str = ..., analog: bool = ..., *args, **kwargs) -> None: ...

class FIR_Filter(BaseFilter):
    def __init__(self, fc, fs, n, Wp, window: str = ..., btype: str = ..., *args, **kwargs) -> None: ...



import numpy as np
import matplotlib.pyplot as plt
from typing import Union,Sequence

def signal_wave(x,
                *y,
                timeAxis=True,
                xlab:Union[str,Sequence]="",
                ylab:Union[str,Sequence]="",
                title:Union[str,Sequence]="",
                savepath=...,
                showfig=True):
    '''显示时域信号波形
    Params
        x:  当timeAxis为True时,x为时间序列
            当timeAxis为False时,x为频率序列
        *y: 当timeAxis为True时,*y为时域信号
            当timeAxis为False时,*y为频率信号
        timeAxis: 为True时x,*y传入时域信号
                为False时x,*y传入频域信号 
        xlab: x轴标签
        ylab: y轴标签
        title: 标题
    '''
    _len = len(y)
    if isinstance(xlab,str):
        xlab = [ xlab for _ in range(_len) ]
    if isinstance(ylab,str):
        ylab = [ ylab for _ in range(_len) ]
    if isinstance(title,str):
        title = [title for _ in range(_len)]
    if isinstance(xlab,Sequence):
        [ xlab.append("") for _ in range(_len - len(xlab)) ]
    if isinstance(ylab,Sequence):
        [ ylab.append("") for _ in range(_len - len(ylab)) ]
    if isinstance(title,Sequence):
        [title.append("") for _ in range(_len - len(title))]
    N = len(x)
    T = x[1]-x[0] if timeAxis else (x[1]-x[0])/N
    t = x if timeAxis else np.arange(N)*T # 时间序列
    _, ax = plt.subplots(len(y),1,squeeze=False)
    for i,u in enumerate(y):
        U = u if timeAxis else np.fft.ifft(u)
        ax[i][0].plot(t,U)
        ax[i][0].set_title(title[i])
        ax[i][0].set_xlabel(xlab[i])
        ax[i][0].set_ylabel(ylab[i])
    
    if savepath != ...:
        plt.savefig(savepath)
    if showfig:
        plt.show()

def signal_fft_wave(x,
                    *y,
                    freqAxis:bool=False,
                    single:bool=True,
                    xlab:Union[str,Sequence]="",
                    ylab:Union[str,Sequence]="",
                    title:Union[str,Sequence]="",
                    savepath=...,
                    showfig=True):
    '''显示信号频域波形
    Params
        x:  当freqAxis为True时,x为频率序列
            当freqAxis为False时,x为时间序列
        *y: 当freqAxis为True时,*y为频率信号
            当freqAxis为False时,*y为时域信号
        freqAxis: 为True时x,*y传入频域信号
                为False时x,*y传入时域信号
        single: 指定是单边波形,还是双边波形
        xlab: x轴标签
        ylab: y轴标签
        title: 标题
    '''  
    _len = len(y)
    if isinstance(xlab,(str)):
        xlab = [ xlab for _ in range(_len) ]
    if isinstance(ylab,str):
        ylab = [ ylab for _ in range(_len) ]
    if isinstance(title,str):
        title = [title for _ in range(_len)]
    if isinstance(xlab,Sequence):
        [ xlab.append("") for _ in range(_len - len(xlab)) ]
    if isinstance(ylab,Sequence):
        [ ylab.append("") for _ in range(_len - len(ylab)) ]
    if isinstance(title,Sequence):
        [title.append("") for _ in range(_len - len(title))]
    N = len(x)
    T = (x[1]-x[0])/N if freqAxis else x[1]-x[0]  # 信号周期
    f = x if freqAxis else np.fft.fftfreq(N, T)   # 信号频率
    _, ax = plt.subplots(len(y),1,squeeze=False)
    for i,u in enumerate(y):
        U = u if freqAxis else np.fft.fft(u)
        ax[i][0].plot(f[:N//2],2.0/N * np.abs(U[0:N//2])) if single else ax[i][0].plot(f,U)
        ax[i][0].set_title(title[i])
        ax[i][0].set_xlabel(xlab[i])
        ax[i][0].set_ylabel(ylab[i])
 
    if savepath != ...:
        plt.savefig(savepath)
    if showfig:
        plt.show()




import os,sys
import re
import json

class Settings(object):
    # set default json file path 
    if getattr(sys, 'frozen', False):
        _abspath = os.path.dirname(os.path.abspath(sys.executable))
    elif __file__:
        _abspath = os.path.dirname(os.path.abspath(__file__)) # current file dir
    _file_name = "settings.json" # default json file name
    
    def __init__(self, settings_path:str=...):
        """INIT SETTINGS
        settings_path: .json file path or directory(include settings.json)
                        default: settings.json file in the directory where the current file resides
        """
        super(Settings, self).__init__()
        # set file path
        self.set_settings_path(settings_path)

        # Just to have objects references
        self.items = {}

        # DESERIALIZE
        self.deserialize() # 从文件中加载json数据,转为dict型数据,赋值给self.items

    def set_settings_path(self, settings_path:str=...):
        """Set json file path
        settings_path: settings.json file path or directory
        """
        settings_path = settings_path if settings_path != Ellipsis else self._abspath
        settings_path = os.path.abspath(os.path.normpath(settings_path)) #规范路径中'//'为'/'

        if settings_path.endswith(".json"):
            self._settings_path = settings_path
        else:
            self._settings_path = os.path.join(settings_path,self._file_name)
        
        if not os.path.exists(self._settings_path):
            sys.exit(f"WARNING: \".json\" file not found! check in the path \"{self._settings_path}\"")
    
    def get_settings_path(self) -> str: 
        """ Get json file path
        """
        return self._settings_path

    def _parse_json(self, json_raw):
        """Remove json comments "//", "/* */"
        json_raw: raw json
        """
        try:
            json_str1 = re.sub(re.compile('(//[\\s\\S]*?\n)'), '', json_raw)
            json_str2 = re.sub(re.compile('(/\*[\\s\\S]*?\*/)'), '', json_str1)
            return json.loads(json_str2)
        except:
            raise Exception("Json file parse fail!")

    def serialize(self, filepath=...):
        """序列化:将字典型数据转为json, 并保存到文件
        """
        if filepath is Ellipsis:
            filepath = self._settings_path
        # WRITE JSON FILE
        with open(filepath, "w", encoding='utf-8') as fd:
            json.dump(self.items, fd, indent=4)

    def deserialize(self, filepath=...):
        """反序列化:从文件中加载json数据, 并转为dict型数据
        """
        if filepath is Ellipsis:
            filepath = self._settings_path
        # READ JSON FILE
        with open(filepath, "r", encoding='utf-8') as fd:
            self.items = self._parse_json(fd.read())




import numpy as np
from ..setting import Settings
from ..io import IO
from .vlp_model import VLP_Model

io = IO()
Args = Settings() # 加载settings.json文件

save_excel_path = f'vlp_data.xlsx'
save_npz_path = f'vlp_data.npz'

def main():
    print("This is a demo about Visible light positioning!")
    print("In an LOS link, one LED light acts as a transmitter to transmit the DC signal, and four PDs act as receivers to receive the optical signal.")
   
    print("Get the training dataset.")
    # 获取训练集数据
    cm = VLP_Model(**Args.items)
    (P1,P2,P3,P4,X,Y,Z) = cm.get_data()

    # 绘图
    P_SUM = P1 + P2 + P3 + P4
    P = P_SUM
    cm.show(P)

    # 构建训练集
    train_in = np.stack([P1,P2,P3,P4],axis=-1)
    train_out = np.stack((X,Y,Z),axis=-1)

    train_dataset = {
        "P1": P1,
        "P2": P2,
        "P3": P3,
        "P4": P4,
        "X": X,
        "Y": Y,
        "Z": Z
    }

    print("Get the testing dataset.")
    # 获取测试集数据
    Args.items['Room']["test_plane"]["gap"] = 0.24  
    cm = VLP_Model(**Args.items)  
    (P1,P2,P3,P4,X,Y,Z) = cm.get_data()  

    P_SUM = P1 + P2 + P3 + P4
    P = P_SUM
    cm.show(P)

    # 构建测试集
    test_in = np.stack([P1,P2,P3,P4],axis=-1)
    test_out = np.stack((X,Y,Z),axis=-1)  

    test_dataset = {
        "P1": P1,
        "P2": P2,
        "P3": P3,
        "P4": P4,
        "X": X,
        "Y": Y,
        "Z": Z
    } 

    # 保存数据
    io.dataset_save_npz(save_npz_path,train_inputs=train_in,train_outputs=train_out,
                                    test_inputs=test_in,test_outputs=test_out)

    io.save_excel(save_excel_path, train_dataset, test_dataset)   

def test_func():
    """检验数据正确性
    """
    print("\nStatistics dataset:")
    print("*"*98)
    (train_inputs, train_outputs), (test_inputs, test_outputs) = io.dataset_load_npz(save_npz_path)
    print(f"train_inputs\t train_outputs\t test_inputs\t test_outputs")
    print(f"{train_inputs.shape}\t {train_outputs.shape}\t {test_inputs.shape}\t {test_outputs.shape}")
    print("train_inputs[:2]:\n",train_inputs[:2],"\ntrain_outputs[:2]:\n",train_outputs[:2])
    print("test_inputs[:2]:\n" ,test_inputs[:2],"\ntest_outputs[:2]:\n",test_outputs[:2])
    print(f"train_inputs: max value = {np.max(train_inputs)},\tmin value = {np.min(train_inputs)}")
    print(f"test_inputs : max value = {np.max(test_inputs)},\tmin value = {np.min(test_inputs)}")
    print("*"*98)

if __name__ == '__main__':
    for a in range(0,1):
        if a == 0:   # 生成数据
            main()
        elif a == 1: # 加载数据 验证数据集
            test_func()




import typing

class VLPError(Exception):
    pass

class VLPTypeError(VLPError):
    """数据类型异常:
    """
    def __init__(self, value, *typeargv:type):
        """数据需为指定的类型
        """
        self._value = value
        self._type = [str(i).split("\'")[1] for i in list(typeargv)]
    
    def __str__(self):
        s = str(type(self._value)).split("\'")[1]
        return repr(f"The value must be of type {self._type}, but this value is of type {s}.")

class VLPSequenceLenError(VLPError):
    """序列长度异常
    """
    def __init__(self,value:typing.Sequence,length:int):
        """序列需固定的长度
        """
        self._value = value
        self._len = length

    def __str__(self) -> str:
        return repr(f"The length of the Sequence has to be {self._len}, but this Sequence has length {len(self._value)}.")

class VLPValueError(VLPError):
    """数据取值异常
    """
    def __init__(self,value,*argv):
        """数据应指定的取值
        """
        self._value = value
        self._argv = list(argv)

    def __str__(self) -> str:
        return repr(f"The value {self._value} is not in {self._argv}.")

class VLPValueRangeError(VLPError):
    """数据取值范围异常
    """
    def __init__(self,value,low,hignt):
        """数据应在指定的范围内,可能不包括端点处的取值
        """
        self._value = value
        self._range = (low,hignt)
        
    def __str__(self) -> str:
        return repr(f"The value {self._value} must be in the range {self._range}.")




import numpy as np
import matplotlib.pyplot as plt
from ..signal import filter,plot

def main():
    print("This is a demo about filter!")
    Fs = 13000 # 采样频率,单位为Hz
    N = 1024  # 采样点数
    t = np.arange(N) * 1/Fs  # 产生一个以采样时间为间隔的时间序列

    print("Generate a cosine signal containing 3KHz, 4KHz and 5KHz.")
    # 产生多个子信号
    f1 = 3000
    f2 = 4000
    f3 = 5000
    Amp = 10

    U1 = Amp * np.cos(2 * np.pi * f1 * t)
    U2 = Amp * np.cos(2 * np.pi * f2 * t)
    U3 = Amp * np.cos(2 * np.pi * f3 * t)

    # 叠加
    U = U1 + U2 + U3 

    # 显示信号的波形
    plot.signal_wave(t,U1,U2,U3,U,xlab='t(s)',ylab=['U1','U2','U3','U'])
    plot.signal_fft_wave(t,U1,U2,U3,U,xlab='f(Hz)',ylab=['U1','U2','U3','U'])

    # 滤波器设计
    Fc = f1 # 滤波频率
    print(f"A {int(Fc/1000)}KHz bandpass filter is used to filter this signal.")
    ft = filter.ButterFilter(Fc,Fs,[200,400],[3,64])        # 巴特沃斯带通滤波器
    # ft = filter.BesselFilter(Fc,Fs,10,400)                  # 贝塞尔滤波器
    # ft = filter.ChebyFilter(Fc,Fs,[300,450],[3,64],cheby=2) # 切比雪夫II型滤波器
    # ft = filter.EllipFilter(Fc,Fs,[300,450],[3,64])         # 椭圆滤波器
    # ft = filter.FIR_Filter(Fc,Fs,30,300,window='hamming')   # Fir滤波器
    ft.show_wave(N)

    # 滤波
    Ulf = ft.filter(U)

    # 滤波信号波形
    plot.signal_wave(t,Ulf,xlab='t(s)',ylab='Ulf')
    plot.signal_fft_wave(t,Ulf,xlab='f(Hz)',ylab='Ulf')

if __name__ == '__main__':
    main()




from ..io import IO as IO
from ..setting import Settings as Settings
from .vlp_model import VLP_Model as VLP_Model
from _typeshed import Incomplete

io: Incomplete
Args: Incomplete
save_excel_path: str
save_npz_path: str

def main() -> None: ...
def test_func() -> None: ...



import numpy as np
from typing import Union,Callable,Sequence
from ..error import VLPSequenceLenError,VLPValueError
from numbers import Number

class LED:
    """LED灯
    """
    def __init__(self,
                fov:Number=90,
                theta:Number=30,
                pos:tuple=(0,0,0),
                Nv:tuple=(0,0,-1),
                origin:tuple= (0,0,0),
                signal:str="None",
                *args,
                **kwargs) -> None:
        """
        fov: LED最大辐射角 (单位: 度)
        theta: LED半功率点半角 (单位: 度)
        pos: LED参考位置 
        Nv: LED法向量
        origin: 原点位置
        signal: LED发射信号,可选"None","DC","AC","AIM",信号序列
        *args和**kwargs:
            signal="None": 无发射信号,无需传参数: 默认led.signal=0
            signal="DC": 发射直流功率,传参详见函数: DC_Power(Pt);
            signal="AC": 发射交流信号(电压信号或电流信号),传参详见函数: AC_Signal(DCbias, Amp, freq, t, signal_func);
            signal="AIM": 强度调制功率信号,传参详见函数: AIM_Signal(Pt, m, freq, t, signal_func)
        """
        self.origin = origin # 原点位置
        self.fov    = fov    # LED最大辐射角
        self.theta  = theta  # LED半功率点半角
        self.pos    = pos    # 参考位置
        self.Nv     = Nv     # LED法向量

        if isinstance(signal, str):
            self.send_signal(signal, *args, **kwargs)  # LED发射信号
        else:
            self.signal = signal  

    @staticmethod
    def get_time_series(fs, npt):
        """获取时间序列
        Params
            fs: 频率
            npt: 序列长度
        Return 
            t:时间序列
        """
        return np.arange(npt) / fs 

    def DC_Signal(self, Pt, *args, **kwargs):
        """直流信号
        Params
            Pt: 直流功率
        Return 
            直流功率
        """
        return Pt

    def AC_Signal(self, DCbias, Amp, freq, t, signal_func:Union[str,np.ufunc]='cos', *args, **kwargs):
        """交流信号
        Params
            DCbias: 直流偏置
            Amp: 交流信号幅值
            freq: 交流信号频率
            t: 时间序列
            signal_func: 交流信号函数('sin', 'cos')
        Return 
            电压或电流信号
        """
        if signal_func not in ('sin', 'cos', np.sin, np.cos):
            raise VLPValueError(signal_func, 'sin', 'cos', np.sin, np.cos)
        
        if isinstance(signal_func, str):
            signal_func = np.sin if signal_func == 'sin' else np.cos

        t = np.array(t)   #时间序列 (npt)
        return DCbias + Amp * signal_func(2 * np.pi * (t * freq + np.random.random())) #交流信号 (npt)

    def AIM_Signal(self, Pt, m, freq, t, signal_func:Union[str,np.ufunc]='cos', *args, **kwargs):
        """强度调制
        Params
            Pt: 功率信号
            m: 调制指数
            freq: 调制信号频率
            t:时间序列
        Return 
            已调制信号
        """
        if signal_func not in ('sin', 'cos', np.sin, np.cos):
            raise VLPValueError(signal_func, 'sin', 'cos', np.sin, np.cos)
        
        if isinstance(signal_func, str):
            signal_func = np.sin if signal_func == 'sin' else np.cos

        return Pt*(1 + m*signal_func(2 * np.pi * (t * freq + np.random.random())))

    def send_signal(self, signal:str, *args, **kwargs):
        """设置内置的LED发射信号
        Params
            signal: LED信号类型
            *args和**kwargs:
                signal="None": 无发射信号,无需传参: 默认led.signal=0;
                signal="DC": 发射直流功率,传参详见函数: DC_Power(Pt);
                signal="AC": 发射交流信号(电压信号或电流信号),传参详见函数: AC_Signal(DCbias, Amp, freq, t, signal_func);
                signal="AIM": 强度调制功率信号,传参详见函数: AIM_Signal(Pt, m, freq, t, signal_func)
        Return
            无
        """
        print("led signal type:", signal)

        if signal not in ('None', 'DC', 'AC', 'AIM'):
            raise VLPValueError('None', 'DC', 'AC', 'AIM')

        if signal == 'DC':    # 直流功率
            self.signal = self.DC_Signal(*args,**kwargs)
        elif signal == 'AC':  # 交流信号
            self.signal = self.AC_Signal(*args,**kwargs)
        elif signal == 'AIM': # 强度调制
            self.signal = self.AIM_Signal(*args,**kwargs)
        else:                 # 无发射信号
            self.signal = 0

    @property
    def signal(self):
        """获取LED发射的信号
        """
        return self._signal

    @signal.setter
    def signal(self, s):
        """设置LED发射的信号
        """
        self._signal = s

    @property
    def origin(self) -> tuple:
        """获取原点位置
        """
        return self._origin

    @origin.setter
    def origin(self, o):
        """设置原点位置
        """
        if not len(o) == 3:
            raise VLPSequenceLenError(o, 3)
        
        self._origin = tuple(o)

    @property
    def pos(self) -> tuple:
        """获取LED灯的相对位置 (偏移位置)
        """
        return self._pos + self.origin
    
    @pos.setter
    def pos(self, pos:tuple):
        """设置LED灯的参考位置
        """
        pos = np.asarray(pos)
        assert len(pos) == 3 or pos.shape[-1] == 3

        if len(pos) == 3:
            pos = np.stack(pos, axis=-1)

        self._pos = pos # (n,3)
        
    @property
    def Nv(self):  
        """获取LED单位法向量  
        """  
        norm = np.linalg.norm(self._Nv, axis=-1) # 求模 
        if self._Nv.ndim > 1:
            norm = np.expand_dims(norm, axis=-1) # (n,1)

        return self._Nv / norm # 单位化

    @Nv.setter
    def Nv(self, n): 
        """设置LED法向量 
        """
        N = np.asarray(n)
        assert len(N) == 3 or N.shape[-1] == 3

        if len(n) == 3:
            N = np.stack(N, axis=-1)
            
        self._Nv = N

    @property
    def fov(self):
        """获取LED灯的最大辐射角 (单位: 度)
        """    
        return self._fov

    @fov.setter
    def fov(self, fov): 
        """设置LED灯的最大辐射角 (单位: 度)
        """
        self._fov = fov

    @property
    def theta(self):  
        """获取半功率点半角 (单位: 度)  
        """  
        return self._theta

    @theta.setter
    def theta(self, t): 
        """设置半功率点半角 (单位: 度)  
        """
        self._theta = t

    @property
    def m(self):
        """获取朗伯发射级数 
        """
        # 角度制转弧度制
        theta_rad = np.radians(self.theta) 
    
        return -np.log(2) / np.log(np.cos(theta_rad))




import numpy as np
from ..error import VLPSequenceLenError as VLPSequenceLenError, VLPValueError as VLPValueError
from numbers import Number
from typing import Union

class LED:
    def __init__(self, fov: Number = ..., theta: Number = ..., pos: tuple = ..., Nv: tuple = ..., origin: tuple = ..., signal: str = ..., *args, **kwargs) -> None: ...
    @staticmethod
    def get_time_series(fs, npt): ...
    def DC_Signal(self, Pt, *args, **kwargs): ...
    def AC_Signal(self, DCbias, Amp, freq, t, signal_func: Union[str, np.ufunc] = ..., *args, **kwargs): ...
    def AIM_Signal(self, Pt, m, freq, t, signal_func: Union[str, np.ufunc] = ..., *args, **kwargs): ...
    def send_signal(self, signal: str, *args, **kwargs): ...
    @property
    def signal(self): ...
    @property
    def origin(self) -> tuple: ...
    @property
    def pos(self) -> tuple: ...
    @property
    def Nv(self): ...
    @property
    def fov(self): ...
    @property
    def theta(self): ...
    @property
    def m(self): ...



from . import demo
from . import error
from . import io
from . import model
from . import setting
from . import signal
from . import tools
from . import vis

from .info import __name__
from .info import __version__
from .info import __author__
from .info import __url__



import scipy.io as scio
import pandas as pd
import numpy as np

class IO:
    """数据保存和加载
    """
    def save_excel(self, fp:str, *data, header=False, index=False):
        """保存为.excel
        """
        dataset = []
        for dat in data:
            dataset.append(pd.DataFrame(dat))
        with pd.ExcelWriter(fp) as xlsx:
            for i,data in enumerate(dataset):
                data.to_excel(xlsx,'Sheet%d'%(i+1),header=header,index=index)

    def load_excel(self, fp:str, sheet:str="Sheet1", header=None):
        """加载excel, 返回numpy类型数据
        """
        # 读取excell数据
        return pd.read_excel(fp,sheet_name=sheet,header=header).to_numpy()  

    def save_npz(self, fp:str, *data:dict):
        """保存为.npz
        """
        dataset = {}
        for dat in data:
            dataset.update(dat)
        np.savez(fp, **dataset)
 
    def load_npz(self, fp, **argkeys):
        """加载npz格式数据, 返回numpy类型数据
        """
        return np.load(fp, **argkeys)

    def save_csv(self, fp:str, *dataFrame:dict):
        """保存为.csv
        """
        data = pd.DataFrame(data)
        data.to_csv(fp) 

    def load_csv(self, fp:str):
        """加载.csv文件, 返回numpy类型数据
        """
        return pd.read_csv(fp).to_numpy()

    def svae_txt(self, fp:str, data:dict):
        """保存为.txt
        """
        data = pd.DataFrame(data)
        np.savetxt(fp, data)
   
    def load_txt(self, fp:str, sep=None, header=None, index=None):
        """加载txt文件, 返回numpy类型数据
        """
        return pd.read_csv(fp, sep=sep, header=header, index_col=index).to_numpy()
    
    def save_mat(self, fp:str, mdict:dict={}):
        """保存为.mat文件
        """
        scio.savemat(fp, mdict)

    def load_mat(self, fp:str, keys:dict={}, default=None):
        """加载.mat文件
        """
        mdict = scio.loadmat(fp)
        output = []
        for key in keys:
            output.append(mdict.get(key, default))
        return output

    def user_save_npz_2d(self, fp:str, train_p=None, train_x=None, train_y=None, test_p=None, test_x=None, test_y=None):
        """用户数据保存2D (npz数据格式)
        """
        np.savez(fp,train_p=train_p,train_x=train_x,train_y=train_y,
                                test_p=test_p,test_x=test_x,test_y=test_y)

    def user_load_npz_2d(self, fp:str):
        """用户数据加载2D (npz数据格式), 返回numpy类型数据
        """
        with np.load(fp) as f:
            train_p,train_x,train_y = f['train_p'],f['train_x'],f['train_y']
            test_p,test_x,test_y = f['test_p'],f['test_x'],f['test_y']
        return (train_p,train_x,train_y), (test_p,test_x,test_y)

    def user_save_npz_3d(self, fp:str, train_p=None, train_x=None, train_y=None, train_z=None, test_p=None, test_x=None, test_y=None, test_z=None):
        """用户数据保存3D (npz数据格式)
        """
        # 数据保存.npz
        np.savez(fp, train_p=train_p, train_x=train_x, train_y=train_y, train_z=train_z,
                            test_p=test_p, test_x=test_x, test_y=test_y, test_z=test_z)

    def user_load_npz_3d(self, fp:str):
        """用户数据加载3D (npz数据格式), 返回numpy类型数据
        """
        with np.load(fp) as f:
            train_p, train_x, train_y, train_z = f['train_p'], f['train_x'], f['train_y'], f['train_z']
            test_p, test_x, test_y, test_z = f['test_p'], f['test_x'], f['test_y'], f['test_z']
        return (train_p,train_x,train_y,train_z), (test_p,test_x,test_y,test_z)

    def dataset_save_npz(self, fp:str, train_inputs=None, train_outputs=None, test_inputs=None, test_outputs=None):
        """保存数据集
        加载: dataset_load_npz(fp)
        """
        np.savez(fp, train_inputs=train_inputs, train_outputs=train_outputs, 
                        test_inputs=test_inputs, test_outputs=test_outputs)

    def dataset_load_npz(self, fp:str):
        """加载数据集, 返回numpy类型数据
        """
        with np.load(fp, allow_pickle=True) as f:
            train_inputs, train_outputs = f['train_inputs'], f['train_outputs'] 
            test_inputs, test_outputs = f['test_inputs'], f['test_outputs']
        return (train_inputs, train_outputs), (test_inputs, test_outputs)




import matplotlib.pyplot as plt
import numpy as np
from ..model import LED
from ..model import PD
from ..model import Room

class VLP_Model:
    def __init__(self,**kwargs):
        # 房间
        room = kwargs["Room"]
        self.room = Room(*room['size'].values(), *room['test_plane'].values(), *room['wall'].values())

        # PD
        pd = kwargs["PD_Info"] 
        pd1 = pd["PD1"]   
        pd2 = pd["PD2"]   
        pd3 = pd["PD3"]   
        pd4 = pd["PD4"]   
        
        # 获取PD位置和法向量
        pd1_pos, pd1_Nv = PD.recv_frame_model(l=pd["l"], alpha=pd1["alpha"], beta=pd1["beta"], center_tp_pos=self.room.get_tp_pos())  
        pd2_pos, pd2_Nv = PD.recv_frame_model(l=pd["l"], alpha=pd2["alpha"], beta=pd2["beta"], center_tp_pos=self.room.get_tp_pos()) 
        pd3_pos, pd3_Nv = PD.recv_frame_model(l=pd["l"], alpha=pd3["alpha"], beta=pd3["beta"], center_tp_pos=self.room.get_tp_pos())  
        pd4_pos, pd4_Nv = PD.recv_frame_model(l=pd["l"], alpha=pd4["alpha"], beta=pd4["beta"], center_tp_pos=self.room.get_tp_pos()) 
        
        self.pd1 = PD(n=pd["n"], fov=pd["fov"], Ar=pd["Ar"], Ts=pd["Ts"], pos=pd1_pos, Nv=pd1_Nv) 
        self.pd2 = PD(n=pd["n"], fov=pd["fov"], Ar=pd["Ar"], Ts=pd["Ts"], pos=pd2_pos, Nv=pd2_Nv)
        self.pd3 = PD(n=pd["n"], fov=pd["fov"], Ar=pd["Ar"], Ts=pd["Ts"], pos=pd3_pos, Nv=pd3_Nv) 
        self.pd4 = PD(n=pd["n"], fov=pd["fov"], Ar=pd["Ar"], Ts=pd["Ts"], pos=pd4_pos, Nv=pd4_Nv) 
        
        # LED
        led = kwargs["LED_Info"] 
        led1 = led["LED1"]
        self.led1 = LED(theta=led["theta"], signal=led["signal"], **led1)

    def get_data(self):
        """计算PD接收LED功率
        outshape: 输出数据形状,默认shape=(n,)(为None不改变形状)
        """
        # PD接收各LED的信号 (l*w*h)
        P_PD1 = self.pd1.recv_led_signal(led=self.led1, room=self.room) 
        P_PD2 = self.pd2.recv_led_signal(led=self.led1, room=self.room)
        P_PD3 = self.pd3.recv_led_signal(led=self.led1, room=self.room) 
        P_PD4 = self.pd4.recv_led_signal(led=self.led1, room=self.room) 

        # 待测点位置
        Xr, Yr, Zr = self.room.get_tp_pos(fmt='c')  #（l*w*h）
        
        return (P_PD1, P_PD2, P_PD3, P_PD4, Xr, Yr, Zr)  
      
    def show(self,z,savepath=...,showfig=True):
        """绘图
        z: 第三维度数据
        """
        xr, yr, _ = self.room.get_tp_grid()
        z = np.reshape(z, xr.shape)  
        ax = plt.axes(projection='3d')
        ax.plot_surface(xr, yr, z, cmap='viridis', edgecolor='none')
        ax.set_xlabel('X')
        ax.set_ylabel('Y')
        ax.set_zlabel('Z')
        if savepath != ...:
            plt.savefig(savepath)
        if showfig:
            plt.show()




import time

def timer(func):
    """装饰器:计时
    """
    def inner(*args, **kwargs):
        start_time = time.time()
        re = func(*args, **kwargs)
        end_time = time.time()
        print("Program execution time: ", end_time - start_time)
        return re
    return inner




import typing
from scipy import signal
import numpy as np
import matplotlib.pyplot as plt
from ..error import VLPValueError

class BaseFilter:
    def __init__(self, fc, fs, D:list=[100,200], G:list=[2,35], btype='bandpass', analog=False, *args, **kwargs) -> None:
        """滤波器基类
        fc: 截止频率
        fs: 采样频率
        D:  [通带到中心的距离(单边通带宽度),阻带到中心的距离(单边阻带宽度)]
        G: 通带增益,阻带衰减
        btype: 滤波器类型,可选['lowpass', 'highpass', 'bandpass', 'bandstop']
        analog: 为True是模拟滤波器,为False是数字滤波器
        """
        if not btype in ['lowpass', 'highpass', 'bandpass', 'bandstop']:
            raise VLPValueError(btype, 'lowpass', 'highpass', 'bandpass', 'bandstop')

        self.Fc = fc # 滤波频率
        self.Fs = fs
        Pd,Sd = D[0],D[1]
        self.Wp = np.array([fc-Pd,fc+Pd])/(fs/2) # 通带频率 
        self.Ws = np.array([fc-Sd,fc+Sd])/(fs/2) # 阻带频率
        self.gp,self.gs = G[0],G[1]
        self.btype = btype
        self.analog = analog

    def show_wave(self, N, whole=False, savepath=..., showfig=True):
        """显示幅频响应波形、相频响应波形
        N: 采样点数
        whole: 单边还是全部
        """
        if self.analog:
            w,h = signal.freqs(self.b,self.a)
        else:
            # 计算H(z)的幅频响应,freqz(b,a,计算点数,采样速率) 
            w,h = signal.freqz(self.b,self.a,np.linspace(0, self.Fs if whole else self.Fs/2, N),whole=whole,fs=self.Fs) 
        _, ax = plt.subplots(2,1)
        ax[0].plot(w,20*np.log10(abs(h)))
        ax[0].set_xlabel('f(Hz)')
        ax[0].set_ylabel('H(dB)')
        ax[0].set_title('f-H')   # 幅频响应波形
        # ax[0].set_ylim(-100,10)
        # angles = np.unwrap(np.angle(h))
        pha = np.angle(h) * 180 / np.pi
        ax[1].plot(w,pha)
        ax[1].set_xlabel('f(Hz)')
        ax[1].set_ylabel('Pha(deg)')
        ax[1].set_title('f-Arc') # 相频响应波形
        if savepath != ...:
            plt.savefig(savepath)
        if showfig:
            plt.show()

    def filter(self, x, lfilter=False):
        """滤波
        x: 滤波信号
        lfilter: 为True使用lfilter滤波;为False使用filtfilt滤波
                 lfilter 滤波后的波形有偏移,filtfilt滤波后的没有偏移
        """
        return signal.lfilter(self.b,self.a,x) if lfilter else signal.filtfilt(self.b,self.a,x)

class ButterFilter(BaseFilter):
    def __init__(self, fc, fs, D:list=[100,200], G:list=[2,35], btype='bandpass', analog=False, *args, **kwargs) -> None:
        """巴特沃斯带通滤波器
        fc: 截止频率
        fs: 采样频率
        D: [通带到中心的距离(单边通带宽度),阻带到中心的距离(单边阻带宽度)]
        G: 通带增益,阻带衰减
        btype: 滤波器类型,可选['lowpass', 'highpass', 'bandpass', 'bandstop']
        analog: 为True是模拟滤波器,为False是数字滤波器
        """
        if not btype in ['lowpass', 'highpass', 'bandpass', 'bandstop']:
            raise VLPValueError(btype, 'lowpass', 'highpass', 'bandpass', 'bandstop')
        
        super(ButterFilter,self).__init__(fc,fs,D,G,btype,analog,*args,**kwargs)
        n, Wn = signal.buttord(self.Wp, self.Ws, self.gp, self.gs,analog=self.analog) # 返回巴特沃斯滤波器最小阶n和截止频率Wn
        self.b, self.a = signal.butter(n, Wn, btype='bandpass', analog=self.analog)   # 返回归一化截止频率Wn的n阶巴特沃斯滤波器的传递函数系数

class ChebyFilter(BaseFilter):
    def __init__(self, fc, fs, D:list=[100,200], G:list=[2,35], btype='bandpass', analog=False, cheby=1, *args, **kwargs):
        """切比雪夫滤波器
        Params
            fc: 截止频率
            fs: 采样率
            D: [通带到中心的距离(单边通带宽度),阻带到中心的距离(单边阻带宽度)]
            G: 通带增益,阻带衰减
            btype: 滤波器类型,可选['lowpass', 'highpass', 'bandpass', 'bandstop']
            analog: 为True是模拟滤波器,为False是数字滤波器
            cheby: 为1时,为切比雪夫I型滤波器;为2时,为切比雪夫II型滤波器
        """
        if not btype in ['lowpass', 'highpass', 'bandpass', 'bandstop']:
            raise VLPValueError(btype, 'lowpass', 'highpass', 'bandpass', 'bandstop')
        if not cheby in [1,2]:
            raise VLPValueError(cheby, 1, 2)

        super(ChebyFilter,self).__init__(fc,fs,D,G,btype,analog,*args,**kwargs)
        if cheby == 1:
            ripple = self.gp
            filter_ord = signal.cheb1ord
            filter_func = signal.cheby1
        else:
            ripple = self.gs
            filter_ord = signal.cheb2ord
            filter_func = signal.cheby2

        n, Wn = filter_ord(self.Wp, self.Ws, self.gp, self.gs,analog=analog) # 返回滤波器最小阶n和截止频率Wn
        self.b, self.a = filter_func(n, ripple, Wn, btype=self.btype, analog=analog)

class EllipFilter(BaseFilter):
    def __init__(self, fc, fs, D:list=[100,200], G:list=[2,35], btype='bandpass', analog=False, *args, **kwargs) -> None:
        """椭圆滤波器
        fc: 截止频率
        fs: 采样频率
        D: [通带到中心的距离(单边通带宽度),阻带到中心的距离(单边阻带宽度)]
        G: 通带增益,阻带衰减
        btype: 滤波器类型,可选['lowpass', 'highpass', 'bandpass', 'bandstop']
        analog: 为True是模拟滤波器,为False是数字滤波器
        """
        if not btype in ['lowpass', 'highpass', 'bandpass', 'bandstop']:
            raise VLPValueError(btype, 'lowpass', 'highpass', 'bandpass', 'bandstop')
        super(EllipFilter,self).__init__(fc,fs,D,G,btype,analog,*args,**kwargs)

        n, Wn = signal.ellipord(self.Wp, self.Ws, self.gp, self.gs,analog=self.analog)    # 返回巴特沃斯滤波器最小阶n和截止频率Wn
        rp,rs = self.gp,self.gs
        self.b, self.a = signal.ellip(n, rp, rs, Wn, btype=self.btype,analog=self.analog) # 返回归一化截止频率Wn的n阶巴特沃斯滤波器的传递函数系数

class BesselFilter(BaseFilter):
    def __init__(self, fc, fs, n, Wn:typing.Union[int,list], btype='bandpass', analog=False, *args, **kwargs) -> None:
        """贝塞尔滤波器
        fc: 截止频率
        fs: 采样频率
        n: 滤波器阶数
        wn: 单边通带宽度
        btype:滤波器类型,可选['lowpass', 'highpass', 'bandpass', 'bandstop']
        analog:为True是模拟滤波器,为False是数字滤波器
        """
        if not btype in ['lowpass', 'highpass', 'bandpass', 'bandstop']:
            raise VLPValueError(btype, 'lowpass', 'highpass', 'bandpass', 'bandstop')
        super(BesselFilter,self).__init__(fc,fs,btype=btype,analog=analog,*args,**kwargs)
        if np.array(Wn).size == 2:
            Wn = np.array([fc-Wn[0],fc+Wn[1]]) / (fs/2)
        else:
            Wn = Wn if btype in ['lowpass', 'highpass'] else np.array([fc-Wn,fc+Wn]) / (fs/2)
        self.b, self.a = signal.bessel(n, Wn, btype=self.btype,analog=self.analog) # 返回归一化截止频率Wn的n阶巴特沃斯滤波器的传递函数系数

class FIR_Filter(BaseFilter):
    def __init__(self, fc, fs, n, Wp, window='hamming', btype='bandpass', *args, **kwargs) -> None:
        """Fir滤波器
        Params
            fc:截止频率
            fs:采样率
            n:滤波器阶数(越大越好)
            Wp:单边通带宽度
            window:窗函数,可选["barthann","bartlett","blackman","blackmanharris","bohman","boxcar","chebwin","cosine",
                        "dpss","exponential","flattop""gaussian","general_cosine","general_gaussian","general_hamming",
                        "hamming","hann","hanning","kaiser","nuttall","parzen","taylor","triang","tukey"]
            btype:滤波器类型,可选[True, False,'lowpass', 'highpass', 'bandpass', 'bandstop']
        """
        if not btype in [True, False, 'lowpass', 'highpass', 'bandpass', 'bandstop']:
            raise VLPValueError(btype, 'lowpass', 'highpass', 'bandpass', 'bandstop')
        super(FIR_Filter,self).__init__(fc,fs,btype=btype,analog=False,*args,**kwargs)
        if np.array(Wp).size == 2:
            Wp = np.array([fc-Wp[0],fc+Wp[1]])/(fs/2)
        else:
            Wp = np.array([fc-Wp,fc+Wp])/(fs/2) if btype in ['bandpass', 'bandstop'] else Wp/ (fs/2) 
        # window = signal.windows.boxcar(n)
        # window = signal.get_window('boxcar',n)
        self.b, self.a = signal.firwin(n+1,Wp,window=window,pass_zero=btype), 1




import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

def draw_position_contrast(pred,true,title="",xlabel="x",ylabel="y",zlabel="z",savepath=...,showfig=True):
    """预测位置和真实位置对比图
    Params
        pred: 预测位置
        true: 真实位置
        title: 标题
        savepath: 保存路径
        showfig: 是否显示图例
    """
    assert len(pred) == len(true) 
    pred = np.asarray(pred)
    # (n,2) or (n,3) -> (2,n) or (3,n)
    if pred.shape[-1] in (2,3):
        pred = np.transpose(pred)
        true = np.transpose(true)
    if pred.shape[0] == 2:
        plt.scatter(*pred, c='none',facecolors='none',edgecolors='r',linewidths=1)
        plt.scatter(*true, c='k',marker='+',linewidths=1)
    elif pred.shape[0] == 3:
        ax = plt.axes(projection='3d')
        ax.scatter3D(*pred, c='none',facecolors='none',edgecolors='r',linewidths=1)
        ax.scatter3D(*true, c='k',marker='+',linewidths=1)
        ax.set_zlabel(zlabel)
    else: 
        plt.scatter(pred, c='none',linewidths=1, facecolors='none',edgecolors='r')
        plt.scatter(true, c='k', marker='+',linewidths=1)
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    plt.title(title)
    plt.legend(['Predicted position','Actual position'])
    if savepath != ...:
        plt.savefig(savepath)
    if showfig:
        plt.show()

def draw_history(history,title="history",savepath=...,showfig=True):
    """绘制网络模型训练过程参数变换图
    Params
        history: 历史值
        title: 标题
        savepath: 保存路径
        showfig: 是否显示图例
    """
    data = pd.DataFrame(history)
    data.plot(figsize=(8,5))
    plt.grid(True)
    plt.title(title)
    plt.xlabel("Iter")
    if savepath != ...:
        plt.savefig(savepath)
    if showfig:
        plt.show()

def draw_error_hist(error,bins=10,rwidth=0.6,title="",savepath=...,showfig=True):
    """绘制误差分布直方图
    """
    fre_tuple = plt.hist(error, bins=bins, rwidth=rwidth, align='left')
    # 绘制统计文本
    [plt.text(y, x, f'{int(x)}', ha='center', va= 'bottom', fontsize=10) for x,y in zip(*fre_tuple[:2])]
    plt.title(title)
    plt.xlabel("Position error")
    plt.ylabel("The number of position error")

    if savepath != ...:
        plt.savefig(savepath)
    if showfig:
        plt.show()
    return fre_tuple

def cdfplot(error,title="",savepath=...,showfig=True):
    """绘制经验误差积累分布图CDF
    Params 
        error: 误差数组
        title: 标题
        savepath: 保存路径
        showfig: 是否显示图例
    """
    assert len(error) > 1
    x = np.asarray(error,dtype=np.float)
    x = x[~np.isnan(x)]     #去除nan
    x = np.sort(x)          #排序
    n = len(x)              
    y = np.arange(1,n+1)/n  #返回1/n步进到1
    notdup = [x[i]-x[i-1]>0 for i in range(1,len(x))]  #返回梯度布尔值
    notdup.append(True) 
    xx = x[notdup]          #去除重复元素
    yy = np.insert(y[notdup],0,0) #在第一位置上插入一个0
    m = np.repeat(np.arange(len(xx)),2)  #在第0轴上复制元素
    xCDF = np.array([-np.inf,*xx[m],np.inf]) #前后延申
    yCDF = np.array([0,0,*yy[m]]) #在第一位置上插入两个0
    
    plt.plot(xCDF,yCDF)  #绘图
    plt.title(title)
    plt.xlabel("Position error")
    plt.ylabel("CDF")

    if savepath != ...:
        plt.savefig(savepath)
    if showfig:
        plt.show()
    return xCDF,yCDF




import os
import numpy as np
import pandas as pd
from typing import Sequence,Union,Callable
from numbers import Number
from ..error import VLPTypeError, VLPSequenceLenError, VLPValueError

class Room:
    """房间
    坐标系的建立:取房间前侧左下墙角为原点, 前侧左下墙角到前侧右下墙角为x轴方向,\
                前侧左下墙角到后侧左下墙角为y轴方向, 前侧左下墙角到前侧左上墙角为z轴方向
            z   ______________________
            |  /|                    /|
            | / |                   / |
            |/__|__________________/  |
            |   |  y              |   | height
            |   | /               |   |
            |   |/                |   |
            |   |_________________|___|
            |  /                  |  /
            | /                   | / width      
          O |/____________________|/______x  
                     length  
    """
    def __init__(self, 
                length:Number, 
                width:Number, 
                height:Number, 
                tp_gap:Union[Number,Sequence]=0.1,
                tp_height:Union[Number,Sequence]=0,
                reflect_wall:Sequence=[], 
                wall_gap:Union[Number,Sequence]=0.1,
                rho:Union[Number,Sequence]=0.5,
                wall_type:str='regular',
                origin:tuple=(0,0,0),
                *args,
                **kwargs) -> None:
        """
        length: 长
        width:  宽
        height: 高
        tp_gap: 定位区域划分网格的间隔  
        tp_height:定位区域距离地面的高度(默认取地面),也可以取[h0,h1]的一个高度范围
        reflect_wall: 存储反射墙壁序列号的列表(0=地面、1-4=前右后左、5=天花板) (在考虑墙壁有反射的情况下,使用该参数)
        wall_gap: 墙壁划分网格的间隔 (在考虑墙壁有反射的情况下,使用该参数)
        rho: 墙壁反射率 (在考虑墙壁有反射的情况下,使用该参数)
        wall_type: 墙壁类型
            可选参数 'regular': 规则墙壁
                    'irregular': 不规则墙壁
                    'rough' : 粗糙墙壁
                    'user'  : 用户自定义墙壁,使用 register_user_wall_angle_func 函数注册
        origin: 原点位置
        """
        self.length = length  # 房间长
        self.width  = width   # 房间宽
        self.height = height  # 房间高
        self.origin = origin  # 原点位置

        self.tp_length = [0, length]  # 定位区域长度范围
        self.tp_width  = [0, width ]  # 定位区域宽度范围
        self.tp_height = tp_height    # 定位区域高度范围

        self.tp_gap = tp_gap          # 测试点平面划分网格的间隔

        # 在考虑墙壁有反射的情况下，使用以下参数
        self.reflect_wall = reflect_wall  # 存储反射墙壁序列号的集合
        self.rho = rho                    # 墙壁反射率
        self.wall_type = wall_type        # 墙壁类型
        self.wall_roughness = 1           # 墙壁粗糙程度:self.wall_type==rough时有效, 该值一般大于1

        self.rp_length = [0, length]      # 前后墙壁反射面长度范围
        self.rp_width  = [0, width ]      # 左右墙壁反射面宽度范围
        self.rp_height = [0, height]      # 前后左右墙壁反射面高度范围

        self.wall_gap = wall_gap          # 墙壁划分网格的间隔
        self.wall_args_path = f"wall_args.xlsx" # 默认墙壁参数地址

    def get_tp_grid(self):
        """获取测试点位置的网格矩阵 (l,w,h)
        """
        tp_gx, tp_gy, tp_gz = self.tp_gap  # 定位区域划分网格的间隔
        tp_l1, tp_l2 = self.tp_length      # 定位区域长度范围
        tp_w1, tp_w2 = self.tp_width       # 定位区域宽度范围
        tp_h1, tp_h2 = self.tp_height      # 定位区域高度范围
        x = np.arange(tp_l1, tp_l2+1e-3, tp_gx)  # 定位区域长度等间隔划分的数组
        y = np.arange(tp_w1, tp_w2+1e-3, tp_gy)  # 定位区域宽度等间隔划分的数组
        z = np.arange(tp_h1, tp_h2+1e-3, tp_gz)  # 定位平面高度等间隔划分的数组
        xr, yr, zr = np.meshgrid(x, y, z, indexing='ij')  # 返回二维网格矩阵
        return np.squeeze([xr, yr, zr])    # 除维数为1的维度 (l,w,h)

    def get_tp_pos(self, fmt='r'):
        """获取测试点位置
        fmt: 输出测试点坐标格式
            fmt='r': pos.shape=(n,3)
            fmt='c': pos.shape=(3,n)
        """
        if not fmt in ('r','c'):
            raise VLPValueError(fmt,'r','c')
        xr, yr, zr = self.get_tp_grid()
        xr, yr, zr = xr.flatten(), yr.flatten(), zr.flatten() # (l,w,h) -> (l*w*h)
        if fmt == 'r':
            pos = np.stack([xr, yr, zr], axis=-1)             # (l*w*h,3)
        else:
            pos = np.reshape([xr,yr,zr], newshape=(3,-1))     # (3,l*w*h)
        return pos

    def get_reflect_wall_pos(self, gap, wall):
        """获取反射单元坐标
        Params
            gap: 墙壁划分网格的间隔
            wall: 第几面墙(0=地板、1-4=前右后左、5=天花板)
        Return
            reflect_pos: 反射点坐标(列表元素个数等于反射单元的个数)
        """
        if wall not in (0,1,2,3,4,5):
            raise VLPValueError(wall,0,1,2,3,4,5)

        gx, gy, gz = gap              # 墙壁网格划分的间隔

        rp_l1, rp_l2 = self.rp_length # 反射墙壁长度范围
        rp_w1, rp_w2 = self.rp_width  # 反射墙壁宽度范围
        rp_h1, rp_h2 = self.rp_height # 反射墙壁高度范围

        if   wall in (0,5):           # 地板或天花板
            x = np.arange(rp_l1+gx/2, rp_l2+1e-3, gx)          # 等间隔取其中点
            y = np.arange(rp_w1+gy/2, rp_w2+1e-3, gy) 
            z = rp_h1 if wall == 0 else rp_h2
            Xw, Yw, Zw = np.meshgrid(x, y, z, indexing='ij')   # 地板和天花板反射平面建立二维网格矩阵 
        elif wall in (1,3):           # 前后墙    
            x = np.arange(rp_l1+gx/2, rp_l2+1e-3, gx)  
            y = rp_w1 if wall == 1 else rp_w2 
            z = np.arange(rp_h1+gz/2, rp_h2+1e-3, gz) 
            Xw, Yw, Zw = np.meshgrid(x, y, z, indexing='ij')   # 前后墙面反射建立二维网格矩阵 
        elif wall in (2,4):           # 左右墙
            x = rp_l2 if wall == 2 else rp_l1
            y = np.arange(rp_w1+gy/2, rp_w2+1e-3, gy) 
            z = np.arange(rp_h1+gz/2, rp_h2+1e-3, gz) 
            Xw, Yw, Zw = np.meshgrid(x, y, z, indexing='ij')   # 左右墙面反射建立二维网格矩阵  (l,w,h)
        else:
            raise ValueError("Wall number must be 0,1,2,3,4,5")
        
        Xw, Yw, Zw  = Xw.flatten(), Yw.flatten(), Zw.flatten() # 打平 (l*w*h)*3
        reflect_pos = np.stack([Xw, Yw, Zw], axis=-1)          # 在最后一维度上堆叠  (l*w*h,3)
        return reflect_pos

    def regular_wall_angle(self, wall, shape):
        """获取规则墙壁反射面元法向量与x轴和z轴的夹角
        Params
            wall: 墙壁序列号
            shape: 角度形状
        Return
            angles: 墙壁反射面元法向量与x轴和z轴的夹角(单位: 度)
        """
        if   wall == 0:  # 地板法向量   Nw = ( 0, 0, 1)
            alpha, beta = np.full(shape=shape,fill_value=  0), np.full(shape=shape,fill_value=  0) # 获取地板参数
        elif wall == 1:  # 前墙法向量   Nw = ( 0, 1, 0)
            alpha, beta = np.full(shape=shape,fill_value= 90), np.full(shape=shape,fill_value= 90) # 获取墙壁参数
        elif wall == 2:  # 右墙法向量   Nw = (-1, 0, 0)
            alpha, beta = np.full(shape=shape,fill_value=180), np.full(shape=shape,fill_value= 90) # 获取墙壁参数
        elif wall == 3:  # 后墙法向量   Nw = ( 0,-1, 0)
            alpha, beta = np.full(shape=shape,fill_value=270), np.full(shape=shape,fill_value= 90) # 获取墙壁参数
        elif wall == 4:  # 左墙法向量   Nw = ( 1, 0, 0)
            alpha, beta = np.full(shape=shape,fill_value=  0), np.full(shape=shape,fill_value= 90) # 获取墙壁参数
        elif wall == 5:  # 天花板法向量 Nw = ( 0, 0,-1)
            alpha, beta = np.full(shape=shape,fill_value=  0), np.full(shape=shape,fill_value=180) # 获取天花板参数
        else:
            raise ValueError("Wall number must be 0,1,2,3,4,5")
        
        angles = np.stack([alpha, beta], axis=-1) # (n,2)
        return angles

    def irregular_wall_angle(self, shape):
        """获取不规则墙壁反射面元法向量与x轴和z轴的夹角
        Params:
            shape: 角度参数形状
        Return 
            angles: 墙壁反射面元法向量与x轴和z轴的夹角(单位: 度)
        """
        alpha  = np.random.uniform(low=0, high=360, size=shape)
        beta   = np.random.uniform(low=0, high=180, size=shape)
        angles = np.stack([alpha, beta], axis=-1)       # (n,2)
        return angles

    def rough_wall_angle(self, wall, shape):
        """获粗糙墙壁反射面元法向量与x轴和z轴的夹角
        Params
            wall: 墙壁序列号
            shape: 角度形状
        Return
            angles: 墙壁反射面元法向量与x轴和z轴的夹角(单位: 度)
        """
        # 获取规则墙壁角度参数
        angle = self.regular_wall_angle(wall, shape)                   # (l1*w1,2)
        alpha, beta = np.split(angle, indices_or_sections=2, axis=-1)  # (l1*w1,2) -> 2*(l1*w1,1)  
        alpha, beta = np.squeeze([alpha, beta])                        # 2*(l1*w1,1) -> 2*(l1*w1)

        # 为规则平滑墙壁添加粗糙[scale越大，墙壁越粗糙]
        alpha = alpha + np.random.normal(scale=self.wall_roughness, size=shape)
        beta  = beta  + np.random.normal(scale=self.wall_roughness, size=shape)
        angles = np.stack([alpha, beta], axis=-1)     # (n,2)
        return angles

    _user_wall_angle_func = None  
    def register_user_wall_angle_func(self, func:Callable[[int,int], Sequence]=None):
        """注册角度函数,用于生成墙壁参数
        Params 
            func: 角度参数形状
        example:
        >>> def func(wall, shape):
        >>>     return ...
        >>> room.register_user_wall_angle_func(func)
        """
        self._user_wall_angle_func = func

    def get_reflect_wall_angle(self, wall, shape, wall_type):
        """获取墙壁反射面元法向量与x轴和z轴的夹角
        Params:
            wall: 墙壁序列号
            shape: 角度参数形状
            wall_type: 墙壁类型,可选参数 'regular', 'irregular', 'rough', 'user'
        Return 
            angles = [[alpha0,beta0],[alpha1,beta1],...] (列表元素个数等于反射单元的个数)
            其中 alpha: 墙壁反射面元法向量与x轴的夹角
                 beta: 墙壁反射面元法向量与z轴的夹角
        """
        if wall_type not in ('regular', 'irregular', 'rough', "user"):
            raise VLPValueError('regular', 'irregular', 'rough', "user")

        if wall_type == 'regular':       # 生成规则墙壁参数
            print(f"Generate regular wall {wall} parameters.")
            angles = self.regular_wall_angle(wall, shape)

        elif wall_type == 'irregular':   # 随机生成墙壁参数
            print(f"Generated irregular wall {wall} parameter.")
            angles = self.irregular_wall_angle(shape)

        elif wall_type == 'rough':       # 生成粗糙墙壁参数
            print(f"Generate rough wall {wall} parameters.")
            angles = self.rough_wall_angle(wall, shape)

        elif wall_type == 'user':        # 使用用户提供函数生成墙壁角度参数
            if self._user_wall_angle_func is None:
                raise Exception("The user is not registered with the wall angle function.")
            else:
                print(f"Generate user function wall {wall} parameters.")
                angles = self._user_wall_angle_func(wall, shape)

        return angles

    def _save_reflect_wall_args(self, fp:str, reflect_pos, angles):
        """保存反射墙壁参数(位置和角度)到excel
        Params 
            fp: 保存路径
            reflect_pos: 墙壁反射面元所在位置
            angle: 获取墙壁反射面元法向量与x轴和z轴的夹角(单位:度)
        """
        if not np.shape(reflect_pos)[-1] == 3:
            raise VLPSequenceLenError(reflect_pos,3)
        if not np.shape(angles)[-1] == 2:
            raise VLPSequenceLenError(angles,2)
        xw, yw, zw  = np.split(reflect_pos, indices_or_sections=3, axis=-1)  # 分割操作
        alpha, beta = np.split(angles, indices_or_sections=2, axis=-1)       
        xw, yw, zw  = np.squeeze([xw, yw, zw])   # 除维数为1的维度  
        alpha, beta = np.squeeze([alpha, beta])               
        wall_args = pd.DataFrame({
            "xw": xw,
            "yw": yw,
            "zw": zw,
            "alpha": alpha,
            "beta": beta
        })
        wall_args.to_excel(fp,index=False,header=True)
        # 保存为npz文件
        # np.savez(fp, alpha=alpha, beta=beta, xw=xw, yw=yw, zw=zw)
    
    def _load_reflect_wall_args(self, fp:str):
        """加载墙壁参数(位置和角度)到excel
        Params 
            fp: 文件路径
        Return
            reflect_pos,angles 反射单元坐标和反射单元角度(单位:度)
        """
        wall_args = pd.read_excel(fp).to_numpy()
        xw, yw, zw, alpha, beta = np.split(wall_args,indices_or_sections=5,axis=-1)
        # 加载npz文件
        # with np.load(fp) as f:
        #     xw,yw,zw,alpha,beta = f['xw'],f['yw'],f['zw'],f['alpha'],f['beta']
        xw, yw, zw  = np.squeeze([xw, yw, zw])   # 除维数为1的维度
        alpha, beta = np.squeeze([alpha, beta])
        reflect_pos = np.stack([xw, yw, zw], axis=-1)
        angles = np.stack([alpha, beta], axis=-1)
        return reflect_pos, angles

    def get_reflect_wall_args(self) -> list:
        """获取墙壁反射面元位置和角度参数
        Params
            无
        Return
            [(reflect_pos, angles),...] 列表元素个数等于反射墙壁的个数
            其中 reflect_pos: 墙壁反射面元位置
                 angles: 墙壁反射面元法向量与x轴和z轴的夹角
        """
        reflect_wall_args = []      # 反射墙壁网格坐标和反射单元角度
        for wall in self.reflect_wall:
            reflect_pos = self.get_reflect_wall_pos(self.wall_gap, wall) # 得到反射点位置 (n,3)
            num_pos = np.shape(reflect_pos)[0]       # 反射点个数

            fpath = self.wall_args_path.format(wall) # 墙壁参数文件路径
            try:
                _, angles = self._load_reflect_wall_args(fpath) # 加载墙壁参数
                num_angle = np.shape(angles)[0]                 # 获取墙壁角度参数的个数
                assert num_pos == num_angle                     # 验证房间尺寸或墙壁划分间隔发生变化

            except FileNotFoundError as e:  # 墙壁参数文件未发现
                print(f"The reflective wall {wall} parameter file was not found.")
                angles = self.get_reflect_wall_angle(wall, shape=num_pos, wall_type=self.wall_type) # 获取墙壁角度参数
                self._save_reflect_wall_args(fpath, reflect_pos, angles)  # 保存参数

            except AssertionError as e:     # 房间尺寸或墙壁划分间隔发生变化
                print(f"The size of the room or the partition interval of the reflective walls changes.")                  
                angles = self.get_reflect_wall_angle(wall, shape=num_pos, wall_type=self.wall_type) # 获取墙壁角度参数
                self._save_reflect_wall_args(fpath, reflect_pos, angles)  # 保存参数

            else:                           # 反射墙壁参数文件加载成功
                print(f"The reflective wall {wall} parameter file is loaded successfully.")
            finally:
                reflect_wall_args.append((reflect_pos, angles))

        return reflect_wall_args

    @property
    def length(self):
        """获取房间长度
        """
        return self._length

    @length.setter
    def length(self, l):
        """设置房间长度
        """
        self._length = l

    @property
    def width(self):
        """获取房间宽度
        """
        return self._width 

    @width.setter
    def width(self, w):
        """设置房间宽度
        """
        self._width = w

    @property
    def height(self):
        """获取房间高度
        """
        return self._height

    @height.setter
    def height(self, h):
        """设置房间高度
        """
        self._height = h

    @property
    def size(self) -> tuple:
        """获取房间大小(length, width, height)
        """
        return (self.length, self.width, self.height)

    @property
    def origin(self) -> tuple:
        """获取原点
        """
        return self._origin

    @origin.setter
    def origin(self, o):
        """设置原点
        """
        if not len(o) == 3:
            raise VLPSequenceLenError(o,3)
        self._origin = tuple(o)

    @property
    def tp_length(self):
        """获取定位区域长度范围 (已偏移)
        """
        return self._tp_length + self.origin[0]
    
    @tp_length.setter
    def tp_length(self, l:Union[Number,Sequence]):
        """设置定位区域长度范围 (未偏移)
        """
        if isinstance(l, Number):               # 定位区域在房间中某一平面上（2D定位）
            l = np.full(shape=(2),fill_value=l) # 填充
        else:
            l = np.array(l)
            assert l.size == 2 and l[0] <= l[1]
        # assert 0 <= l[0] <= l[1] <= self.length # 对取 [l1,l2] 的一个长度范围限制（3D定位）
        self._tp_length = l

    @property
    def tp_width(self):
        """获取定位区域宽度范围 (已偏移)
        """
        return self._tp_width + self.origin[1]
    
    @tp_width.setter
    def tp_width(self, w:Union[Number,Sequence]):
        """设置定位区域宽度范围 (未偏移)
        """
        if isinstance(w, Number):               # 定位区域在房间中某一平面上（2D定位）
            w = np.full(shape=(2),fill_value=w) # 填充
        else:
            w = np.array(w)
            assert w.size == 2 and w[0] <= w[1]
        # assert 0 <= w[0] <= w[1] <= self.width # 对取 [w1,w2] 的一个宽度范围限制（3D定位）
        self._tp_width = w

    @property
    def tp_height(self):
        """获取定位区域高度范围 (已偏移)
        """
        return self._tp_height + self.origin[2]
    
    @tp_height.setter
    def tp_height(self, h:Union[Number,Sequence]):
        """设置定位区域高度范围 (未偏移)
        """
        if isinstance(h, Number):               # 定位区域在房间中某一高度上（2D定位）
            h = np.full(shape=(2),fill_value=h) # 填充
        else:
            h = np.array(h) 
            assert h.size == 2 and h[0] <= h[1]
        # assert 0 <= h[0] <= h[1] <= self.height # 对取 [h1,h2] 的一个高度范围限制（3D定位）
        self._tp_height = h

    @property
    def tp_gap(self):
        """获取定位区域划分网格的间隔
        """
        return self._tp_gap
    
    @tp_gap.setter
    def tp_gap(self, gap:Union[Number, Sequence]):
        """设置定位区域划分网格的间隔
        """
        gap = np.asarray(gap)
        if not gap.size in (1,3):
            raise VLPValueError(gap.size,1,3)
        if gap.size == 1:
            gap = np.full(shape=(3), fill_value=gap) # 填充
            
        # 对定位区域划分间隔限制
        assert np.all(gap >= 0) 
        self._tp_gap = gap

    @property
    def reflect_wall(self) -> list:
        """获取存储反射墙壁序列号的列表
        """
        reflect_wall = set(self._reflect_wall) # 去除重复元素
        return np.array(list(reflect_wall))
    
    @reflect_wall.setter
    def reflect_wall(self, wall:Sequence):
        """设置存储反射墙壁序列号的列表
        """
        for w in wall:
            if not isinstance(w, int):
                raise VLPTypeError(w, int)
            if w not in (0,1,2,3,4,5):
                raise ValueError("Wall number must be 0,1,2,3,4,5")
        self._reflect_wall = wall

    @property
    def rho(self):
        """获取墙壁反射率
        """
        return self._rho
    
    @rho.setter
    def rho(self, n:Union[Number, Sequence]):
        """设置墙壁反射率
        """
        n = np.asarray(n)
        if not n.size in (1, self.reflect_wall.size): # 实际反射墙壁个数与反射面折射率个数需相等
            raise VLPValueError(n.size, 1, self.reflect_wall.size)
        if n.size == 1:
            n = np.full(shape=(self.reflect_wall.size), fill_value=n) # 填充

        self._rho = n

    @property
    def Aw(self):
        """获取墙壁反射面元面积
        """
        gx, gy, gz = self.wall_gap # 墙壁网格划分的间隔(x,y,z轴方向)
        self._Aw = []
        for wall in self.reflect_wall:
            if wall in (0,5):   # 地板或天花板
                self._Aw.append(gx * gy) 
            elif wall in (1,3): # 前后墙
                self._Aw.append(gx * gz)
            else:               # 左右墙
                self._Aw.append(gy * gz) 
                
        return self._Aw

    @property
    def rp_length(self):
        """获取前后墙反射面长度范围 (已偏移)
        """
        return self._rp_length + self.origin[0]
    
    @rp_length.setter
    def rp_length(self, l:Sequence):
        """设置前后墙反射面长度范围 (未偏移)
        """
        l = np.asarray(l)
        if not l.size == 2:
            raise VLPSequenceLenError(l.size,2)
        
        assert l[0] <= l[1]
        # assert 0 <= l[0] <= l[1] <= self.length # 对取l的反射面长度限制
        self._rp_length = l

    @property
    def rp_width(self):
        """获取左右墙反射面宽度范围 (已偏移)
        """
        return self._rp_width + self.origin[1]
    
    @rp_width.setter
    def rp_width(self, w:Sequence):
        """设置左右墙反射面宽度范围 (未偏移)
        """
        w = np.asarray(w)
        if not w.size == 2:
            raise VLPSequenceLenError(w.size,2)
        
        assert w[0] <= w[1]
        # assert 0 <= w[0] <= w[1] <= self.width # 对取h的反射面的宽度限制
        self._rp_width = w

    @property
    def rp_height(self):
        """获取前后左右墙反射面高度范围 (已偏移)
        """
        return self._rp_height + self.origin[2]
    
    @rp_height.setter
    def rp_height(self, h:Sequence):
        """设置前后左右墙反射面高度范围 (未偏移)
        """
        h = np.asarray(h)
        if not h.size == 2:
            raise VLPSequenceLenError(h.size,2)
        
        assert h[0] <= h[1]
        # assert 0 <= h[0] <= h[1] <= self.height # 对取h的反射面高度限制
        self._rp_height = h

    @property
    def wall_gap(self):
        """获取墙壁划分网格的间隔 
        """
        return self._wall_gap
    
    @wall_gap.setter
    def wall_gap(self, gap:Union[Number, Sequence]):
        """设置墙壁划分网格的间隔
        """
        gap = np.asarray(gap)
        if not gap.size in (1,3):
            raise VLPValueError(gap.size,1,3)
        if gap.size == 1:
            gap = np.full(shape=(3), fill_value=gap)
        # 墙壁划分反射单元网格的间隔限制
        assert np.all(gap >= 0)
        self._wall_gap = gap 

    @property
    def wall_args_path(self):
        """获取墙壁参数保存地址
        """
        (filename,ext) = os.path.splitext(self._wall_args_path) # 分割文件名和后缀名
        fp = filename+'{}'+ext
        return fp
    
    @wall_args_path.setter
    def wall_args_path(self, fp:str):
        """设置墙壁参数保存地址
        """
        self._wall_args_path = fp




from ..model import LED as LED, PD as PD, Room as Room
from _typeshed import Incomplete

class VLP_Model:
    room: Incomplete
    pd1: Incomplete
    pd2: Incomplete
    pd3: Incomplete
    pd4: Incomplete
    led1: Incomplete
    def __init__(self, **kwargs) -> None: ...
    def get_data(self): ...
    def show(self, z, savepath=..., showfig: bool = ...) -> None: ...



from ..error import VLPSequenceLenError as VLPSequenceLenError, VLPTypeError as VLPTypeError, VLPValueError as VLPValueError
from _typeshed import Incomplete
from numbers import Number
from typing import Callable, Sequence, Union

class Room:
    wall_type: Incomplete
    wall_roughness: int
    def __init__(self, length: Number, width: Number, height: Number, tp_gap: Union[Number, Sequence] = ..., tp_height: Union[Number, Sequence] = ..., reflect_wall: Sequence = ..., wall_gap: Union[Number, Sequence] = ..., rho: Union[Number, Sequence] = ..., wall_type: str = ..., origin: tuple = ..., *args, **kwargs) -> None: ...
    def get_tp_grid(self): ...
    def get_tp_pos(self, fmt: str = ...): ...
    def get_reflect_wall_pos(self, gap, wall): ...
    def regular_wall_angle(self, wall, shape): ...
    def irregular_wall_angle(self, shape): ...
    def rough_wall_angle(self, wall, shape): ...
    def register_user_wall_angle_func(self, func: Callable[[int, int], Sequence] = ...): ...
    def get_reflect_wall_angle(self, wall, shape, wall_type): ...
    def get_reflect_wall_args(self) -> list: ...
    @property
    def length(self): ...
    @property
    def width(self): ...
    @property
    def height(self): ...
    @property
    def size(self) -> tuple: ...
    @property
    def origin(self) -> tuple: ...
    @property
    def tp_length(self): ...
    @property
    def tp_width(self): ...
    @property
    def tp_height(self): ...
    @property
    def tp_gap(self): ...
    @property
    def reflect_wall(self) -> list: ...
    @property
    def rho(self): ...
    @property
    def Aw(self): ...
    @property
    def rp_length(self): ...
    @property
    def rp_width(self): ...
    @property
    def rp_height(self): ...
    @property
    def wall_gap(self): ...
    @property
    def wall_args_path(self): ...



import os, shutil
from setuptools import setup, find_packages

#移除构建的build文件夹
CUR_PATH = os.path.dirname(os.path.abspath(__file__))
path = os.path.join(CUR_PATH, 'build')
if os.path.isdir(path):
    print('INFO del dir ', path) 
    shutil.rmtree(path)

setup(
    name = 'vlppy', #应用名
    version = '0.1.15',  #版本号
    description = 'Construction of visible light positioning model', 
    packages = find_packages(),  #包括在安装包内的Python包
    include_package_data = True, #启用清单文件MANIFEST.in,包含数据文件
    # exclude_package_data = {'docs':['1.txt']},  #排除文件
    install_requires = [#自动安装依赖
        'numpy>=1.19.0',
        'matplotlib>=3.5.0',
        'scipy>=1.8.0',
        'pandas>=1.3.0',
    ],
    author = 'yangwuju',
    author_email = '1424134319@qq.com',
    url = 'https://gitee.com/yangwuju/vlppy.git',
)





from ..error import VLPSequenceLenError as VLPSequenceLenError, VLPValueRangeError as VLPValueRangeError
from .led import LED as LED
from .room import Room as Room
from numbers import Number
from typing import Union

class PD:
    def __init__(self, n: Number = ..., fov: Number = ..., Ar: Number = ..., Ts: Number = ..., pos: tuple = ..., Nv: tuple = ..., origin: tuple = ..., *args, **kwargs) -> None: ...
    @staticmethod
    def recv_frame_model(l, alpha, beta, center_tp_pos: tuple): ...
    @staticmethod
    def recv_surface_model(r: Union[Number, tuple], alpha, beta, center_tp_pos: tuple): ...
    @staticmethod
    def recv_hemisphere_model(r, l, alpha, center_tp_pos: tuple): ...
    def recv_led_radiation_intensity(self, led: LED): ...
    def recv_los_led_signal(self, led: LED): ...
    def recv_nlos_led_signal(self, led: LED, room: Room): ...
    def recv_led_signal(self, led: LED, room: Room): ...
    def shot_noise(self, P, R: float = ..., Ibg: float = ..., I2: float = ..., B: float = ..., *argv, **kwargv): ...
    def thermal_noise(self, Tk: int = ..., G: int = ..., eta: int = ..., B: float = ..., I2: float = ..., I3: float = ..., gamma: float = ..., gm: int = ..., *argv, **kwargv): ...
    def recv_noise_signal(self, P, *argv, **kwargv): ...
    def SNR(self, P, R, dB: bool = ..., *argv, **kwargv): ...
    @property
    def origin(self) -> tuple: ...
    @property
    def pos(self): ...
    @property
    def Nv(self): ...
    @property
    def fov(self): ...
    @property
    def Ar(self): ...
    @property
    def Ts(self): ...
    @property
    def n(self): ...
    @property
    def G_Con(self): ...



import numpy as np
from numbers import Number
from typing import Union
from .led import LED
from .room import Room
from ..error import VLPSequenceLenError,VLPValueRangeError

class PD:
    """光电探测器(PD)
    """
    def __init__(self,
                n:Number=1.5,
                fov:Number=60,
                Ar:Number=1e-4,
                Ts:Number=1,
                pos:tuple=(0,0,0),
                Nv:tuple=(0,0,1),
                origin:tuple= (0,0,0),
                *args,
                **kwargs) -> None:
        """ 
        n: 透镜的折射率
        fov: 接收视场角 (单位: 度) [0,90]
        Ar: PD检测器的有效接收面积(单位:平方米)
        Ts: 光学滤波器的增益
        pos: PD参考位置
        Nv: PD接收面的法向量
        origin: 原点位置
        """
        self.n   = n
        self.fov = fov
        self.Ar  = Ar
        self.Ts  = Ts

        self.origin = origin
        self.pos    = pos
        self.Nv     = Nv

    @staticmethod
    def recv_frame_model(l, alpha, beta, center_tp_pos:tuple):
        """接收机模型:PD支架模型(倾斜PD和中心参考点位置关系)
        Params
            l: 倾斜PD到中心水平PD(参考点)的长度(单位:米)
            alpha: PD摆放角(单位:度)
            beta: PD倾斜面倾斜角(单位:度)
            center_tp_pos: 水平中心测试点位置
        Return
            pos: 倾斜PD位置
            Nv: 倾斜PD的法向量
        """
        if len(center_tp_pos) == 3:
            x0, y0, z0 = center_tp_pos
        else:
            x0, y0, z0 = np.split(center_tp_pos, indices_or_sections=3, axis=-1)
        x0, y0, z0 = np.squeeze([x0, y0, z0]) # 除维数为1的维度  
        
        # 角度制转弧度制
        alpha_rad, beta_rad = np.radians([alpha, beta])

        # 倾斜PD和中心参考点位置关系
        xr = x0 + l * np.cos(alpha_rad) * np.cos(beta_rad) 
        yr = y0 + l * np.sin(alpha_rad) * np.cos(beta_rad)
        zr = z0 + l * np.sin(beta_rad)
        
        # alpha_rad+np.pi: PD的方位角和PD倾斜面的摆放角反向
        # beta_rad: PD的法向量与z轴的夹角和PD倾斜面的倾斜角相等
        Nx = np.cos(alpha_rad+np.pi) * np.sin(beta_rad)
        Ny = np.sin(alpha_rad+np.pi) * np.sin(beta_rad)
        Nz = np.cos(beta_rad)

        if len(center_tp_pos) == 3:
            pos = np.asarray([xr,yr,zr])
            Nv = np.asarray([Nx,Ny,Nz]) 
        else:
            pos = np.stack([xr,yr,zr], axis=-1) # 在最后一维度上堆叠
            Nv = np.stack([Nx,Ny,Nz], axis=-1)
        
        return pos, Nv

    @staticmethod
    def recv_surface_model(r:Union[Number,tuple], alpha, beta, center_tp_pos:tuple):
        """接收机模型:安全帽椭球面(倾斜PD和中心参考点位置关系)
        Params
            r: 曲球面半径(单位:米)
            alpha: PD摆放角(单位:度)
            beta: 椭球球心与PD连线与z轴夹角(单位:度)
            center_tp_pos: 椭球面顶部中心测试点位置
        Return
            pos: 倾斜PD位置
            Nv: 倾斜PD的法向量
        """
        r = np.asarray(r)
        assert r.size in (1,3)
        a, b, c = np.full(shape=(3), fill_value=r) if r.size == 1 else r

        # 角度制转弧度制
        alpha_rad, beta_rad = np.radians([alpha, beta]) 

        if not 0 <= beta_rad <= 90: # 限定
            raise VLPValueRangeError(beta, 0, 90)
        
        if len(center_tp_pos) == 3:
            x0, y0, z0 = center_tp_pos
        else:
            x0, y0, z0 = np.split(center_tp_pos, indices_or_sections=3, axis=-1)
        x0, y0, z0 = np.squeeze([x0, y0, z0]) # 除维数为1的维度  
        
        # 倾斜PD位置
        xr = x0 + a * np.cos(alpha_rad) * np.sin(beta_rad)
        yr = y0 + b * np.sin(alpha_rad) * np.sin(beta_rad)
        zr = z0 - c * (1 - np.cos(beta_rad))

        # alpha_rad: PD的方位角和PD倾斜面的摆放角相等
        # beta_rad: 椭球球心与PD连线与z轴夹角
        Nx = 1/a * np.cos(alpha_rad) * np.sin(beta_rad)
        Ny = 1/b * np.sin(alpha_rad) * np.sin(beta_rad)
        Nz = 1/c * np.cos(beta_rad)

        # PD位置与法向量
        if len(center_tp_pos) == 3:
            pos = np.asarray([xr,yr,zr])
            Nv = np.asarray([Nx,Ny,Nz]) 
        else:
            pos = np.stack([xr,yr,zr], axis=-1)
            Nv = np.stack([Nx,Ny,Nz], axis=-1)

        return pos, Nv

    @staticmethod
    def recv_hemisphere_model(r, l, alpha, center_tp_pos:tuple):
        """接收机模型:半球面模型(倾斜PD和中心参考点位置关系)
        Params
            r: 曲球面半径(单位:米)
            l: 倾斜PD到顶部中心(参考点)的弧长(单位:米)   
            alpha:倾斜PD方位角(单位:度) 
            center_tp_pos: 测试点位置 
        Return
            pos: 倾斜PD位置
            Nv: 倾斜PD的法向量
        """
        # 倾斜PD的方位角和倾斜角
        beta_rad = l/r              # 椭球球心与PD连线与z轴夹角(PD倾斜角)
        beta = np.degrees(beta_rad) # 弧度制转角度制
        pos, Nv = PD.recv_surface_model(r, alpha, beta, center_tp_pos)

        return pos, Nv

    def recv_led_radiation_intensity(self, led:LED):
        """计算LED辐射强度(LOS链路)
        Params
            led: LED实例化
        Return 
            LOS链路LED辐射强度
        """
        Vt_r = self.pos - led.pos                                           # LED到接收点的方向向量 (l*w*h,3)                                      # LED到PD的方向向量 (l*w*h,3)
        d = np.linalg.norm(Vt_r, axis=-1)                                   # LED到接收点的方向向量的模 (l*w*h)
        cos_led_t_angle_rad_rad = np.sum(Vt_r*led.Nv, axis=-1) / d          # LED发射角的余弦值 (l*w*h)
        led_t_angle_rad = np.arccos(cos_led_t_angle_rad)                    # LED发射角 (l1*w1)   
        led_t_angle = np.degrees(led_t_angle_rad)                           # 弧度制转为角度制 (l1*w1)

        cos_led_t_angle_rad = np.abs(cos_led_t_angle_rad)                   # 在LED辐射角内,解决辐射角大于90°,角余弦值为负值情况
        cos_led_t_angle_rad = np.where((led_t_angle>=0) & (led_t_angle<=led.fov), cos_led_t_angle_rad, 0) # 满足LED灯照射条件

        R0 = ((led.m + 1)  / (2 * np.pi)) * cos_led_t_angle_rad_rad**led.m  # 辐射强度模型 (l*w*h)

        """ y(t) = x(t) * h(t)
        H0 = H0.flatten() #打平 (l,w) -> (l*w)
        return np.array([np.convolve(h,led.send_signal) for h in H0])  #卷积 (l*w,npt)
              <=> np.array([h * led.send_signal for h in H0])          #相乘 (l*w,npt) 
        """
        # 以下是为了兼容交直流信号
        R0 = np.expand_dims(R0, -1)              # 在最后一维扩充一个维度 (l*w*h) -> (l*w*h,1)  <=> H0.reshape(l*w*h,1) <=> H0[...,np.newaxis]
        radiation_intensity = R0 * led.signal    # LED辐射强度[广播机制 (l*w*h,1) * (npt) -> (l*w*h,npt) * (l*w*h,npt) = (l*w*h,npt)]

        return np.squeeze(radiation_intensity)   # (l*w*h,npt).如果npt==1,则返回形状为(l*w*h)

    def recv_los_led_signal(self, led:LED):
        """计算PD接收LED信号(LOS链路):
        Params
            led: LED实例化
        Return 
            LOS链路中PD探测器接收LED信号
        """
        #⭐ LED -> PD
        # LED发射端
        Vt_r = self.pos - led.pos                                   # LED到PD的方向向量 (l*w*h,3)
        d = np.linalg.norm(Vt_r, axis=-1)                           # 求模:LED到PD的方向向量的模 (l*w*h)
        cos_led_t_angle_rad = np.sum(Vt_r*led.Nv, axis=-1) / d      # LED发射角的余弦值 (l*w*h)
        led_t_angle_rad = np.arccos(cos_led_t_angle_rad)            # LED发射角 (l1*w1)   
        led_t_angle = np.degrees(led_t_angle_rad)                   # 弧度制转为角度制 (l1*w1)

        cos_led_t_angle_rad = np.abs(cos_led_t_angle_rad)           # 在LED辐射角内,解决辐射角大于90°,角余弦值为负值情况
        cos_led_t_angle_rad = np.where((led_t_angle>=0) & (led_t_angle<=led.fov), cos_led_t_angle_rad, 0) # 满足LED灯照射条件

        # PD接收端
        Vr_t = -Vt_r # PD到LED的方向向量 (l*w*h,3)
        cos_pd_r_angle_rad = np.sum(Vr_t*self.Nv, axis=-1) / d      # PD接收角的余弦值 (l*w*h) 
        pd_r_angle_rad = np.arccos(cos_pd_r_angle_rad)              # PD接收器入射角的大小(弧度制) (l*w*h)
        pd_r_angle = np.degrees(pd_r_angle_rad)                     # 弧度制转为角度制 (l*w*h)

        cos_pd_r_angle_rad = np.abs(cos_pd_r_angle_rad)             # 解决PD入射角大于90°,入射角余弦值为负值情况
        cos_pd_r_angle_rad = np.where((pd_r_angle>=0) & (pd_r_angle<=self.fov), cos_pd_r_angle_rad, 0)  # 满足PD入射条件

        # 计算信道增益
        H0 = (((led.m + 1) * self.Ar) / (2 * np.pi * d**2)) * cos_led_t_angle_rad**led.m * cos_pd_r_angle_rad * self.Ts * self.G_Con # (l*w*h)

        """ y(t) = x(t) * h(t)
        H0 = H0.flatten() # 打平 (l,w) -> (l*w,)
        return np.array([np.convolve(h,led.send_signal) for h in H0])  # 卷积 (l*w,npt)
              <=> np.array([h * led.send_signal for h in H0])          # 相乘 (l*w,npt) 
        """
        # 以下是为了兼容交直流信号
        H0 = np.expand_dims(H0, axis=-1)  # 在最后一维扩充一个维度 (l*w*h) -> (l*w*h,1)  <=> H0.reshape(l*w*h,1) <=> H0[...,np.newaxis]
        recv_signal = H0 * led.signal     # PD接收信号 [广播机制 (l*w*h,1) * (npt) -> (l*w*h,npt) * (l*w*h,npt) = (l*w*h,npt)]

        return np.squeeze(recv_signal)    # (l,w,h,npt).如果npt==1,则返回形状为(l,w,h)
      
    def recv_nlos_led_signal(self, led:LED, room:Room):
        """计算PD接收LED一次反射信号(NLOS链路):
        Params
            led: LED实例化
            room: Room实例化
        Return 
            NLOS链路中PD探测器接收LED一次反射信号
        """
        # 对PD坐标张量扩充维度,以便使用广播
        pd_pos = np.expand_dims(self.pos, axis=-2)  # (l*w*h,3) -> (l*w*h,1,3) 

        H0 = 0  #反射(LOS)链路信道增益
        # 遍历每一面反射墙壁参数:(l1*w1,3),(l1*w1,2)
        for i, (w_pos, angle) in enumerate(room.get_reflect_wall_args()):
            w_alpha, w_beta = np.split(angle, indices_or_sections=2, axis=-1)   # (l1*w1,2) -> (l1*w1,1) + (l1*w1,1)
            w_alpha, w_beta = np.squeeze([w_alpha, w_beta])                     # (l1*w1,1) -> (l1*w1)

            # 墙壁反射面元法向量与x轴和z轴的夹角,角度制转弧度制
            w_alpha_rad, w_beta_rad = np.radians([w_alpha, w_beta]) # (l1*w1)
            
            # 墙壁反射面元法向量
            Nx = np.cos(w_alpha_rad) * np.sin(w_beta_rad)
            Ny = np.sin(w_alpha_rad) * np.sin(w_beta_rad)
            Nz = np.cos(w_beta_rad)

            Nw = np.stack([Nx,Ny,Nz], axis=-1)             # 墙壁反射点法向量 (l1*w1,3)
            Nw_norm = np.linalg.norm(Nw, axis=-1)          # 墙壁反射点法向量模 (l1*w1)
            if Nw.ndim > 1:
                Nw_norm = np.expand_dims(Nw_norm, axis=-1) # (l1*w1,1)
            Nw = Nw/Nw_norm                                # 墙壁反射点的单位法向量 (l1*w1,3)

            #⭐ LED -> 墙壁反射面元
            # LED发射端
            Vt_w = w_pos - led.pos              # LED灯到墙壁反射面元的方向向量 (l1*w1,3)
            d1 = np.linalg.norm(Vt_w, axis=-1)  # 求模:LED灯到墙壁反射面元的距离 (l1*w1)
            mark1 = np.all(Vt_w==0, axis=-1)    # 异常点处理:判断墙壁反射面元和LED是否重合(方向向量是否为零向量) (l1*w1)
            d1 = np.where(mark1, 1, d1)         # 异常点处理:为使cos_led_t_angle_rad分母不为0,让墙壁反射面元和LED重合时的距离不为0 (l1*w1)
            
            cos_led_t_angle_rad = np.sum(Vt_w*led.Nv, axis=-1) / d1    # LED发射角的余弦值 (l1*w1)  
            led_t_angle_rad = np.arccos(cos_led_t_angle_rad)           # LED发射角 (l1*w1)   
            led_t_angle = np.degrees(led_t_angle_rad)                  # 弧度制转为角度制 (l1*w1)  

            cos_led_t_angle_rad = np.abs(cos_led_t_angle_rad)          # 解决LED发射角大于90°,发射角余弦值为负值情况
            cos_led_t_angle_rad = np.where((led_t_angle>=0) & (led_t_angle<=led.fov), cos_led_t_angle_rad, 0) # 满足LED灯照射条件

            # 墙壁反射面元接收端
            Vw_t = -Vt_w     # 墙壁反射面元到LED灯的方向向量 (l1,w1,3)
            cos_wall_r_angle_rad = np.sum(Vw_t*Nw, axis=-1) / d1                             # 墙壁反射面元入射角 (l1*w1)
            cos_wall_r_angle_rad = np.where(cos_wall_r_angle_rad>0, cos_wall_r_angle_rad, 0) # 满足墙壁反射条件 (入射角小于90)

            #⭐墙壁反射面元 -> PD
            # 墙壁反射面元发射端
            Vw_r = pd_pos - w_pos              # 墙壁反射面元到PD的方向向量  (l*w*h,1,3)-(l1*w1,3) -> (l*w*h,l1*w1,3)
            d2 = np.linalg.norm(Vw_r, axis=-1) # 求模:墙壁反射面元到PD之间的距离 (l*w*h,l1*w1)
            mark2 = np.all(Vw_r==0, axis=-1)   # 异常点处理:判断墙壁反射面元和参考点是否重合(方向向量是否为零向量) (l*w*h,l1*w1)
            d2 = np.where(mark2, 1, d2)        # 异常点处理:为使cos_wall_t_angle_rad分母不为0,让墙壁反射面元和参考点重合时的距离不为0

            cos_wall_t_angle_rad = np.sum(Vw_r*Nw, axis=-1) / d2                             # 墙壁反射面元发射角 (l*w*h,l1*w1)
            cos_wall_t_angle_rad = np.where(cos_wall_t_angle_rad>0, cos_wall_t_angle_rad, 0) # 满足墙壁反射条件 (发射角小于90)
    
            # PD接收端
            Vr_w = -Vw_r                                             # PD到墙壁反射面元的方向向量 (l*w*h,l1*w1,3)
            cos_pd_r_angle_rad = np.sum(Vr_w*self.Nv, axis=-1) / d2  # PD接收角的余弦值 (l*w*h,l1*w1)
            pd_r_angle_rad = np.arccos(cos_pd_r_angle_rad)           # PD接收器入射角(弧度制) (l*w*h,l1*w1)
            pd_r_angle = np.degrees(pd_r_angle_rad)                  # 弧度制转为角度制 (l*w*h,l1*w1)
            
            cos_pd_r_angle_rad = np.abs(cos_pd_r_angle_rad)          # 解决PD入射角大于90°,入射角余弦值为负值情况
            cos_pd_r_angle_rad = np.where((pd_r_angle>=0) & (pd_r_angle<=self.fov), cos_pd_r_angle_rad, 0)  # 满足PD入射条件

            rho = room.rho[i]   # 墙壁反射率
            Aw  = room.Aw[i]    # 墙壁反射面元面积

            # 反射(NLOS)子链路信道增益  
            Hw = (((led.m + 1) * self.Ar) / (2 * np.pi * d1**2 * d2**2)) * rho * Aw * cos_led_t_angle_rad**led.m * cos_wall_r_angle_rad * cos_wall_t_angle_rad *  cos_pd_r_angle_rad * self.Ts * self.G_Con  # (l*w*h,l1*w1)
            Hw = np.sum(Hw, axis=-1)   # 反射(NLOS)链路信道增益求和 (l*w*h,l1*w1) -> (l*w*h)

            # 对所有反射墙壁(NLOS)链路信道增益求和  
            H0 += Hw                   # (l*w*h)

        # 墙壁一次反射(NLOS)链路信道增益
        H0 = np.expand_dims(H0,-1)     # 在最后一维扩充一个维度 (l*w*h) -> (l*w*h,1)  <=> H0.reshape(l*w*h,1) <=> H0[...,np.newaxis]
        recv_signal = H0 * led.signal  # PD接收信号[广播机制 (l*w*h,1) * (npt) -> (l*w*h,npt) * (l*w*h,npt) = (l*w*h,npt)]
        return np.squeeze(recv_signal) # 去除维度为1的维数
    
    def recv_led_signal(self, led:LED, room:Room):
        """计算PD接收LED信号(LOS+NLOS链路):
        Params
            led: LED实例化
            room: Room实例化
        Return 
            LOS+NLOS链路中PD探测器接收LED信号
        """
        recv_los_signal  = self.recv_los_led_signal (led=led)            # PD接收直射链路信号
        recv_nlos_signal = self.recv_nlos_led_signal(led=led, room=room) # PD接收一次反射链路信号
        return recv_los_signal + recv_nlos_signal

    def shot_noise(self, P, R=0.62, Ibg=2e-9, I2=0.562, B=100e6, *argv, **kwargv):
        """获取PD散粒噪声功率
        Params
            P: PD接收功率(W)
            R: PD响应度(A/W)
            Ibg: 暗电流(A)
            I2: 噪声带宽系数
            B: 噪声带宽(Hz)
        Return 
            PD散粒噪声功率
        """
        q = 1.6e-19  # 电子电荷量(C)
        return 2*q*R*P*B + 2*q*Ibg*I2*B
    
    def thermal_noise(self, Tk=300, G=10, eta=112, B=100e6, I2=0.562, I3=0.0868, gamma=1.5, gm=30, *argv, **kwargv):
        """获取PD热噪声功率
        Params
            Tk: 热力学温度:Tk = T + 273.15°C(K)
            G: 开环电压增益
            eta: PD每单位面积上固定电容(pF/cm2)
            B: 噪声带宽(Hz)
            I2: 噪声带宽系数 
            I3: 噪声带宽系数 
            gamma: 信道噪声系数
            gm: 跨导(mS)
        Return
            PD热噪声功率
        """
        k = 1.38e-23  # 玻尔兹曼常数
        eta *= 1e-8   # 单位换算(1pF/cm2 = 1e-8F/m2)
        gm *= 1e-3    # 单位换算(1mS = 1e-3S)

        return 8*np.pi*k*Tk/G*eta*self.Ar*I2*B**2 + \
            16*np.pi**2*k*Tk*gamma/gm*eta**2*self.Ar**2*I3*B**3

    def recv_noise_signal(self, P, *argv, **kwargv):
        """接收噪声信号功率
        """
        return self.shot_noise(P=P, *argv, **kwargv) + self.thermal_noise(*argv, **kwargv)

    def SNR(self, P, R, dB=True, *argv, **kwargv):
        """信噪比
        Params
            P: PD接收LOS+NLOS链路的光功率
            R: PD的响应度
            dB: 信噪比dB表示
        Return
            信噪比
        """
        P_noise = self.shot_noise(P=P, R=R, *argv, **kwargv) + self.thermal_noise(*argv, **kwargv)
        snr = (R*P)**2 / P_noise
        return 10*np.log10(snr) if dB else snr

    @property
    def origin(self) -> tuple:
        """设置原点位置
        """
        return self._origin

    @origin.setter
    def origin(self, o):
        """获取原点位置
        """
        if not len(o) == 3:
            raise VLPSequenceLenError(o,3)
        
        self._origin = tuple(o) 

    @property
    def pos(self):
        """获取PD的相对位置 (偏移位置)
        """
        return self._pos + self.origin
    
    @pos.setter
    def pos(self, pos:tuple):
        """设置PD的参考位置
        """
        pos = np.asarray(pos)
        assert len(pos) == 3 or pos.shape[-1] == 3

        if len(pos) == 3:
            xr,yr,zr = pos
            pos = np.stack([xr,yr,zr], axis=-1)

        self._pos = pos # (L*w*h,3)

    @property
    def Nv(self):  
        """获取PD单位法向量  
        """  
        norm = np.linalg.norm(self._Nv, axis=-1) # 求模 (n)
        if self._Nv.ndim > 1:
            norm = np.expand_dims(norm, axis=-1) # (n,1)

        return self._Nv / norm # 单位化 (n,3)
    
    @Nv.setter
    def Nv(self, n):
        """设置PD法向量  
        """  
        N = np.asarray(n)
        assert len(n) == 3 or n.shape[-1] == 3

        if len(n) == 3:
            Nx,Ny,Nz = N
            N = np.stack([Nx,Ny,Nz], axis=-1)

        self._Nv = N # (n,3)

    @property
    def fov(self):
        """获取接收视场角 (单位: 度)
        """    
        return self._fov

    @fov.setter
    def fov(self, fov): 
        """设置接收视场角 (单位: 度)
        """
        # if not 0 <= fov <= 90:
        #     raise VLPValueRangeError(fov,0,90)
        self._fov = fov

    @property
    def Ar(self):   
        """获取PD检测器的接收面积(单位:平方米)
        """ 
        return self._Ar

    @Ar.setter
    def Ar(self, a): 
        """设置PD检测器的接收面积(单位:平方米)
        """
        self._Ar = a

    @property
    def Ts(self):   
        """获取光学滤波器的增益 
        """ 
        return self._Ts

    @Ts.setter
    def Ts(self, ts): 
        """设置光学滤波器的增益 
        """
        self._Ts = ts

    @property
    def n(self): 
        """获取透镜的折射率
        """   
        return self._n

    @n.setter
    def n(self, n): 
        """设置透镜的折射率
        """
        self._n = n

    @property
    def G_Con(self):
        """获取聚光器增益
        """
        # 角度制转弧度制
        fov_rad = np.radians(self.fov)

        return self.n**2 / np.sin(fov_rad)**2  




from typing import Sequence, Union

def signal_wave(x, *y, timeAxis: bool = ..., xlab: Union[str, Sequence] = ..., ylab: Union[str, Sequence] = ..., title: Union[str, Sequence] = ..., savepath=..., showfig: bool = ...): ...
def signal_fft_wave(x, *y, freqAxis: bool = ..., single: bool = ..., xlab: Union[str, Sequence] = ..., ylab: Union[str, Sequence] = ..., title: Union[str, Sequence] = ..., savepath=..., showfig: bool = ...): ...



from _typeshed import Incomplete

class Settings:
    items: Incomplete
    def __init__(self, settings_path: str = ...) -> None: ...
    def set_settings_path(self, settings_path: str = ...): ...
    def get_settings_path(self) -> str: ...
    def serialize(self, filepath=...) -> None: ...
    def deserialize(self, filepath=...) -> None: ...



from _typeshed import Incomplete

class IO:
    def save_excel(self, fp: str, *data, header: bool = ..., index: bool = ...): ...
    def load_excel(self, fp: str, sheet: str = ..., header: Incomplete | None = ...): ...
    def save_npz(self, fp: str, *data: dict): ...
    def load_npz(self, fp, **argkeys): ...
    def save_csv(self, fp: str, *dataFrame: dict): ...
    def load_csv(self, fp: str): ...
    def svae_txt(self, fp: str, data: dict): ...
    def load_txt(self, fp: str, sep: Incomplete | None = ..., header: Incomplete | None = ..., index: Incomplete | None = ...): ...
    def save_mat(self, fp: str, mdict: dict = ...): ...
    def load_mat(self, fp: str, keys: dict = ..., default: Incomplete | None = ...): ...
    def user_save_npz_2d(self, fp: str, train_p: Incomplete | None = ..., train_x: Incomplete | None = ..., train_y: Incomplete | None = ..., test_p: Incomplete | None = ..., test_x: Incomplete | None = ..., test_y: Incomplete | None = ...): ...
    def user_load_npz_2d(self, fp: str): ...
    def user_save_npz_3d(self, fp: str, train_p: Incomplete | None = ..., train_x: Incomplete | None = ..., train_y: Incomplete | None = ..., train_z: Incomplete | None = ..., test_p: Incomplete | None = ..., test_x: Incomplete | None = ..., test_y: Incomplete | None = ..., test_z: Incomplete | None = ...): ...
    def user_load_npz_3d(self, fp: str): ...
    def dataset_save_npz(self, fp: str, train_inputs: Incomplete | None = ..., train_outputs: Incomplete | None = ..., test_inputs: Incomplete | None = ..., test_outputs: Incomplete | None = ...): ...
    def dataset_load_npz(self, fp: str): ...
